
/*
 * Some code fragments illustrating the conversion of Basic.java to C/C++.
 */


//typedef void (*StatementHandler)();
//StatementHandler jump_table[100];



#include <stdio.h>
#include <stdarg.h>  // error handler

#include "bool.h"
#include "lex.h"
#include "token.h"
#include "SymbolTable.h"
#include "Stack.h"
#include "expr.h"
#include "stmt.h"
#include "struct.h"

void goto_line(int n);


#define MAX_LINE_LEN 238

// program buffer
char prog_buf[4096];

// array with pointers to lines
char *lines[1000];

// global symbol table
SymbolTable symbols;
Stack stack;

// current line number
int lineno;

// pointer to current line
const char *line;

bool quit = false;
bool end = false;




/*
 * Simple error handler.
 * In the original BBC BASIC, the default error handler is a BASIC routine!
 */
void error(int num, const char *msg, ...)
{
  va_list arglist;

  va_start(arglist, msg);
  vfprintf(stderr, msg, arglist );
  va_end(arglist);

  fprintf(stderr, " at line %d\n", lineno + 1);

  // print detailed error position
  fprintf(stderr, "%s\n", line);
  for(const char *p = line; p < pos; p++)
    fputc(' ', stderr);
  fprintf(stderr, "^\n");

  exit(num);
}



// ---- get and set variable values ----

void getVariable(const char *name, Object *val)
{
  Object *ret = symbols.get(name);
  if(!ret)
    error(1, "Unknown variable: %s", name);  // need var args!
  *val = *ret; // copy it
  //memcpy(val, ret, sizeof(Object));
}


/*
 *  Assigns the value pointed to by val to the variable pointed to by dest.
 *  If the type cannot be casted, a type mismatch error is generated.
 */
void assign(Object *dest, Object *val)
{
  switch(dest->type) {
  case INTEGER:
    dest->intValue = val->toInt();
    break;
  case FLOAT:
    dest->floatValue = val->toFloat();
    break;
  case STRING:
    dest->stringValue = val->asString();
  }
}


void assign(const char *name, Object *val)
{
//   if(isString(name) && !val->isString())
//     error(6, "Type mismatch: string needed");

//   if(!isString(name) && val->isString())
//     error(6, "Type mismatch: number needed");
  Object *dest = symbols.lookup_or_create(name);
  assign(dest, val);

  // -> symbols.put() kann man abschaffen...
}


// if we don't wanna do this in SymbolTable.cpp...

Object *lookup_or_create(const char *name)
{
  Object *ret = symbols.get(name);
  if(ret)
    return ret;

  // put dummy value
  Object val;
  ret = symbols.put(name, val);

  return ret;
}


// ---- parser helper functions: ----

/*
 *  Skips current token if it matches with given token,
 *  generates an error with the given code and message otherwise.
 */
void expect(int token, int code, const char *msg)
{
  if(ttype != token)
    error(code, msg);
  nextToken();
}


void expect(int token)
{
  if(ttype != token)
    error(1, "Syntax error: %s expected, found %s", tokenToString(token),
	  tokenToString(ttype));
  nextToken();
}


void jump_to(int l, char *p)
{
  lineno = l;    // set line number
  lex_start(p);  // set lexer position
  nextToken();   // load first token
  line = lines[lineno];  // set line pointer
}


/*
 * Interprets the current statement.
 * The first token of the statement must already be in ttype.
 * This method can also be called inside an IF ... THEN <statement> ELSE ...
 * structure. It doesn't check if the remaining token is EOL or ':'.
 *
 * The return value is provided to indicate "block end statements"
 * that end a block (like ENDIF, ENDWHILE).
 */
int interpret_statement()
{
  switch(ttype) {
  case TT_REM:
  case TT_DATA:
  case TT_DEF:
    // ignore rest of line
    // HACK: pretend we were at end of line and return
    ttype = TT_EOL;
    return 0;
  case TT_IDENTIFIER:
    // assignment
    Object val, *var;
    var = symbols.lookup_or_create(sval);
    nextToken();
    expect('=');  // raise "unknown command" error instead...
    expression(&val);
    assign(var, &val);
    break;
  case TT_EOL:  // empty line
  case ':':     // or empty statement
    return 0;
    // these statements end blocks -> return to enclosing function
  case TT_ENDCASE:
  case TT_OTHERWISE:
  case TT_WHEN:
    return 1;

    /*
     * Branch to statement handlers
     */

  case TT_REPEAT:
    exec_repeat();  // exec_repeat_until();
    break;
  case TT_UNTIL:
    exec_until();
    break;
  case TT_FOR:
    exec_for();  //exec_for_next();
    break;
  case TT_NEXT:
    exec_next();
    break;
  case TT_IF:
    exec_if();
    break;
  case TT_ELSE:
    exec_else();
    break;
  case TT_ENDIF:
    //exec_endif();
    // currently, ENDIFs are just ignored
    nextToken();
    break;
  case TT_WHILE:
    exec_while();
    break;
  case TT_ENDWHILE:
    exec_endwhile();
    break;
  case TT_PRINT:
    exec_print();
    break;
  default:
    error(1, "Unexpected token at start of statement");
  }

  // ensure_eos();

  return 0;
}


// A statement must end with end of line, a colon or ELSE

void ensure_eos()
{
  if(ttype != TT_EOL && ttype != ':' && ttype != TT_ELSE)
    error(1, "Syntax error: end of statement expected");
}


void interpret_line(const char *l)
{
  //printf("--> %s\n", l);
  line = l;
  lex_start(line);
  do {
    nextToken();
    interpret_statement();
  }
  while(ttype == ':');

  if(ttype != TT_EOL)
    error(1, "End of line expected");
}


/*
 * Enters interactive mode.
 */
void interactive_mode()
{
  char line[255];

  //printf("Interactive mode\n");
  do {
    printf("\nREADY.\n");
    if(!fgets(line, sizeof(line), stdin))
      return;
    interpret_line(line);
  }
  while(!quit);
}


/*
 * Loads a named program file.
 * The lines are stored in prog_buf.
 */
void load(const char *filename)
{
  FILE *c;
  char *p;
  int n = 0, len;

  // open file
  if(!(c = fopen(filename, "r"))) {
    fprintf(stderr, "File not found: %s\n", filename);
    return;
  }

  p = prog_buf;
  while(fgets(p, MAX_LINE_LEN, c)) {
    len = strlen(p);
    // strip trailing newline
    if(p[len - 1] == '\n')
      p[--len] = '\0';
    lines[n++] = p;
    p += len + 1;
  }
  lines[n] = NULL;

  fclose(c);
}


/*
 *  Lists the current program to stdout, with preceding physical line numbers.
 */
void list()
{
  for(int n = 0; lines[n]; n++)
    printf("%d %s\n", n+1, lines[n]);
}


/*
 *  Interprets the current program. Exits when the program has ended
 *  (regularly or due to an error).
 */
void run()
{
  goto_line(0);
  do {

    interpret_statement();

    if(ttype == TT_EOL) {
      // next line
      if(lines[lineno + 1])
	goto_line(lineno + 1);
      else
	return;  // end
    }
    else if(ttype == ':')
      nextToken();  // skip colons
    // else statement has jumped to beginning of a statement (hopefully)
  }
  while(true);
}


/*
 *  Main program:
 *  If there is a filename given, load that program and run it, then quit.
 *  Else print a welcome message and enter interactive mode.
 */
int main(int argc, char **argv)
{
  if(argc > 1) {
    load(argv[1]);
    //list();
    run();
    //exec_lvar();
  }
  else {
    printf("ZAP BASIC version 0.10 (C) Sebastian Mennicke 2003\n");
    interactive_mode();
  }
}
