/*
 *  Implementation of structure statements with a stack.
 *  The stack holds pointers to "StackFrames".
 *  Disadvantage: we have to define a StackFrame class for each structure.
 */

#include "Object.h"
#include "SymbolTable.h"
#include "error.h"
#include "bool.h"
#include "lex.h"
#include "token.h"
#include "expr.h"
#include "arg.h"

#include "basic.h"

void goto_line(int n);


/*
 *  StackFrame types
 */
enum { SF_IF, SF_FOR, SF_REPEAT, SF_WHILE };


class StackFrame
{
public:
  int type;
};


class ForFrame : public StackFrame
{
public:
  ForFrame(Object *v, float l, float s, int n, char *p) :
    var(v), limit(l), step(s), lineno(n), pos(p) { type = SF_FOR; }

  Object *var;
  float limit;
  float step;
  int lineno;
  char *pos;
};


class RepeatFrame : public StackFrame
{
public:
  RepeatFrame(int n, char *p) :
    lineno(n), pos(p) { type = SF_REPEAT; }

  int lineno;
  char *pos;
};


class WhileFrame : public StackFrame
{
public:
  WhileFrame(int n) :
    lineno(n) { type = SF_WHILE; }

  int lineno;
};



void exec_repeat()
{
  stack.push(new RepeatFrame(lineno, pos));
  nextToken();
}


void exec_until()
{
  nextToken();  // skip UNTIL
  RepeatFrame *top = (RepeatFrame *) stack.peek();
  if(top->type != SF_REPEAT)
    error(1, "Not in a REPEAT loop");
  if(float_expression() == 0)
    jump_to(top->lineno, top->pos);
  else {
    stack.pop();
    delete top;
  }
}


void exec_for()
{
  Object *var;
  float limit, step;

  nextToken();
  if(ttype != TT_IDENTIFIER)
    error(1, "Identifier expexted after FOR");
  if(isString(sval))
    error(6, "Type mismatch: numeric variable needed");
  var = symbols.lookup_or_create(sval);
  nextToken();
  expect('=', 1, "Missing = in FOR statement");

  Object val;
  expression(&val);
  assign(var, &val);
  expect(TT_TO, 1, "Missing TO");
  limit = float_expression();
  if(ttype == TT_STEP) {
    nextToken();
    // distinguish between int and float...
    step = float_expression();
    if(step == 0)
      error(35, "The step cannot be zero");
  }
  else
    step = 1;

  stack.push(new ForFrame(var, limit, step, lineno, pos));
}


void exec_next()
{
  nextToken();
  ForFrame *top = (ForFrame *) stack.peek();
  if(top->type != SF_FOR)
    error(1, "Not in a FOR loop");

  // var points to the loop variable

  if(top->var->isInteger())
    top->var->intValue += (int) top->step;
  else
    top->var->floatValue += top->step;
  //assign(top->var, &val);
  if(top->var->toFloat() <= top->limit)
    jump_to(top->lineno, top->pos);
  else
    delete (ForFrame *) stack.pop();
}


void gotoElseOrEndif()
{
  int level = 1;
  do {
    lineno++;  // next line
    lex_start(lines[lineno]);
    nextToken();
    if(ttype == TT_IF) level++;
    else if(ttype == TT_ENDIF) level--;
  }
  while(!(ttype == TT_ENDIF && level == 0 || ttype == TT_ELSE && level == 1));
  nextToken();  // should be TT_EOL, checked at end of interpret()
}


void exec_if()
{
  nextToken();
  float cond = float_expression();
  expect(TT_THEN, 1, "Missing THEN");
  if(ttype == TT_EOL) {
    // multi-line if statement
    if(cond == 0)
      gotoElseOrEndif();
  }
  else {
    // single line if statement
  }
}


void exec_else()
{
  int level = 1;
  do {
    lineno++;  // next line
    lex_start(lines[lineno]);
    nextToken();
    if(ttype == TT_IF) level++;
    else if(ttype == TT_ENDIF) level--;
  }
  while(!(ttype == TT_ENDIF && level == 0));
  nextToken();
}



void gotoEndwhile()
{
  int level = 1;
  do {
    lineno++;  // next line
    lex_start(lines[lineno]);
    nextToken();
    if(ttype == TT_WHILE) level++;
    else if(ttype == TT_ENDWHILE) level--;
  }
  while(!(ttype == TT_ENDWHILE && level == 0));
  nextToken();  // should be TT_EOL, checked at end of interpret()
}


void exec_while()
{
  nextToken();  // skip WHILE
  if(float_expression() == 0)
    gotoEndwhile();
  else
    stack.push(new WhileFrame(lineno));
}


void exec_endwhile()
{
  nextToken();  // skip ENDWHILE
  // pop frame, a new one will be pushed in the next loop
  WhileFrame *top = (WhileFrame *) stack.pop();
  if(top->type != SF_WHILE)
    error(1, "Not in a WHILE loop");
  goto_line(top->lineno - 1);  // exec WHILE again
  delete top;
}
